-- Ghost definitions
ghost_dumps  = { "ghost_JXQ.dump" }

-- Timing options
sync_mode    = "ingame"
display_mode = "ingame"
offset_mode  = "room"

ghost_hitbox = false
samus_hitbox = false
enemy_hitbox = false
shot_hitbox = false

enemy_hp = false
global_map = true
show_delays = true
show_status = false

-- Graphics options
own_color = "white"
ghost_color = { "red", "blue", "green", "orange" }
enemy_color = "yellow"
shot_color = "purple"
map_pos = { 0xbe, 0, 0x42, 0x1f }

-- These require gd
ghost_gfx    = { 1, 2, 3 } -- nil to turn off. Array to specify individually
pose_info = { { "poses_red.png", 0x18, 0x20 } }
minimap = "minimap.png"

-- Main parameters end here
if ghost_gfx or minimap then require "gd" end

-- Variables that must be saved in savestates
last_igframe = 0 -- Used to find out if this frame should be skipped
last_room    = 0
last_transition = { 0, 0, 0} -- Used for room sync: room, frame, igframe
fcount = 1 -- Frames since script start. Used in place of movie.framecount

-- Frames: For both real frames and ingame frames, the first frame of
-- is number 1. All frames before this are labeled 0.

-- Map area offsets in the global map, corrected for snes offsets
region_pos = { [0]={x=-2,y=0}, [1]={x=-5,y=21}, [2]={x=26,y=41}, [3]={x=32,y=-10}, [4]={x=23,y=21}, [5]={x=-5,y=0}, [6]={x=20,y=0} }

-- The main function. It is run every frame.
function main()
	frame = framecount()
	room = get_room()

	handle_input()
	if display() then
		--update_screen()
		snes9x.speedmode("normal")
	else snes9x.speedmode("maximum") end
	fcount = fcount +1
	last_igframe = igframe()
	if room ~= last_room then
		last_transition = { last_room, frame, igframe() }
	end
	last_room = room
end

function update_screen()
	room = get_room()
	if global_map then init_global_map() end
	for index, ghost in ipairs(ghosts) do repeat
		-- Hack: We have an off by one error due to recorder and player
		-- running at different stages. This results in the ghosts being
		-- one frame ahead of the rest (I think that's the reason anyway).
		-- Fix it by tweaking the framecount manually:
		local sframe  = syncframe()-1
		local gframe  = sync2frame(sframe, ghost)
		local osframe = offset(sframe, ghost)
		local ogframe = osframe and sync2frame(osframe, ghost)

		if(ogframe and room == ghost.data[ogframe].room) then
			if ghost_hitbox then draw_ghost_hitbox(ghost, ogframe) end
			if ghost.gfx then draw_ghost_gfx(ghost, ogframe) end
		end
		if global_map and gframe then draw_global_map_ghost(ghost, gframe) end
		if show_delays and osframe then draw_delay(ghost,osframe-sframe, index) end
	until true end
	if samus_hitbox then draw_own_hitbox() end
	if shot_hitbox then draw_shot_hitbox() end
	if enemy_hitbox then draw_enemy_hitbox() end
	if enemy_hp then draw_enemy_hp() end
	if global_map then draw_global_map_samus() end
	if show_status then draw_status() end
end

-- Main data table. An array of tables for each ghost.
-- Each ghost has the following entries:
-- Offset: scalar, number of sync frames to shift by
-- Framenums: array, translates from sync frames to real frames
-- Transitions: dictionary of { from, to } -> array of real frames,
--              from and two might have to be encoded as strings or
--              put into one number (need 5 bytes, and double should
--              provide enough
-- Data: frame -> x, y, region, roomx, roomy, pose
-- Gfx: Scalar. Which of the gfx sets this ghost uses
ghosts = {}
pose_data = {}

function init()
	set_sync(sync_mode)
	set_display(display_mode)
	set_offset(offset_mode)
	for i,filename in ipairs(ghost_dumps) do
		local ghost = readghost(filename)
		ghost.gfx = tabnum(ghost_gfx, i)
		ghost.color = tabnum(ghost_color, i)
		table.insert(ghosts, ghost)
	end
	if ghost_gfx then for i,info in ipairs(pose_info) do
		table.insert(pose_data, read_pose(info))
	end end
	if minimap then mini = gd.createFromPng(minimap) end
	-- Set up saves
	savestate.registersave(function() return last_igframe, last_room, last_transition[1], last_transition[2], last_transition[3], fcount end)
	savestate.registerload(function(_,a,b,c,d,e,f) last_igframe, last_room, last_transition[1], last_transition[2], last_transition[3], fcount = a, b, c, d, e, f end)
	gui.register(update_screen)
end

function readghost(filename)
	local ghost = {}
	local pattern = "^%s*(%d+)%s+([%w_]+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)"
	ghost.offset = 0
	ghost.data = {}
	for line in io.lines(filename) do
		local e = {}
		local reg, rx, ry, px, py, hx, hy, p, ig
		_, _, ig, _, reg, rx, ry, px, py, hx, hy, p = string.find(line, pattern)
		e.room = join_room(tonumber(reg), tonumber(rx), tonumber(ry))
		e.igframe, e.posx, e.posy, e.hitx, e.hity, e.pose =
			tonumber(ig), tonumber(px), tonumber(py), tonumber(hx),
			tonumber(hy), tonumber(p)
		table.insert(ghost.data,e)
	end
	-- Fix the broken first part
	fix_data(ghost.data)
	-- Now extract the information we want
	ghost.syncframes = build_sync(ghost)
	ghost.transitions = build_transitions(ghost)
	return ghost
end

function tabnum(something, index)
	if type(something) == "table" then
		return something[index]
	else
		return something
	end
end

function make_frame2sync(key)
	if key == "real" or key == "realtime" then
		return function(rframe, ghost) return rframe end
	elseif key == "game" or key == "ingame" then
		return function(rframe, ghost) return ghost.data[rframe].igframe end
	end
end
function make_sync2frame(key)
	return function(frame, ghost) return ghost.syncframes[frame] end
end

function framecount()
	return (movie.framecount() or framecount-1)+1
end
function igframe()
	return 3600*memory.readbyte(0x7E09DE)+60*memory.readbyte(0x7E09DC)+memory.readbyte(0x7E09DA)
end
function make_syncframe(key)
	if key == "real" or key == "realtime" then
		return framecount
	elseif key == "game" or key == "ingame" then
		return igframe
	end
end
function make_display(key)
	if key == "real" or key == "realtime" then
		return function() return true end
	elseif key == "game" or key == "ingame" then
		return function() return igframe() ~= last_igframe end
	end
end
function make_offset(key)
	if key == "none" then
		return function(sframe, ghost) return sframe end
	elseif key == "room" then
		return function(sframe, ghost)
			local last_room, last_frame = get_last_trans()
			local t = find_transition(ghost, last_room, room, last_frame)
			return t and sframe - last_frame + t
		end
	end
end
function make_last_trans(key)
	if key == "real" or key == "realtime" then
		return function() return last_transition[1], last_transition[2] end
	elseif key == "game" or key == "ingame" then
		return function() return last_transition[1], last_transition[3] end
	end
end

function nsync(ghost) return #ghost.syncframes end
function nframe(ghost) return #ghost.data end

function set_sync(mode)
	frame2sync = make_frame2sync(mode)
	sync2frame = make_sync2frame(mode)
	syncframe  = make_syncframe(mode)
	get_last_trans = make_last_trans(mode)
end

function set_display(mode)
	display    = make_display(mode)
end

function set_offset(mode)
	offset     = make_offset(mode)
end

function fix_data(data)
	local i, n = 1, #data
	while i < n and data[i].igframe ~= 0 do i = i+1 end
	for j = i-1,1,-1 do
		data[j].igframe, data[j].region, data[j].roomx,
		data[j].roomy, data[j].posx, data[j].posy, data[j].hitx, data[j].hity,
		data[j].pose = 0,0,0,0,0,0,0,0,0
	end
end

-- This assumes that igframe changes by a step of +1 only
function build_sync(ghost)
	local res = {}
	local last = 0
	for i = 1, nframe(ghost) do
		if frame2sync(i, ghost) ~= last then
			table.insert(res,i)
			last = frame2sync(i,ghost)
		end
	end
	return res
end

function get_room()
	return join_room(memory.readword(0x7e079f), memory.readword(0x7e07a1), memory.readword(0x7e07a3))
end
-- math.pow because SHIFT can't access bits above 32
function join_room(region, x, y)
	return region * math.pow(2,32) + SHIFT(x, -16) + y
end
function split_room(room)
	return math.floor(room / math.pow(2,32)),AND(SHIFT(room,16),0xffff),AND(room,0xffff)
end

function appendtrans(trans, from, to, frame)
	if not trans[from] then trans[from] = {} end
	if not trans[from][to] then trans[from][to] = {} end
	table.insert(trans[from][to], frame)
end

function gettrans(trans, from, to)
	return trans[from] and trans[from][to]
end

function build_transitions(ghost)
	local res = {}
	local last = 0
	for i = 1, nframe(ghost) do
		local new = ghost.data[i].room
		if last ~= new then
			appendtrans(res,last,new,i)
			last = new
		end
	end
	return res
end

-- Find the transition closest to sync frame "near_sync"
-- May return nil
function find_transition(ghost, from, to, near_sync)
	if near_sync < 1 then return 0 end -- handle last frame
	local t = gettrans(ghost.transitions, from, to)
	local ns = math.min(near_sync,nsync(ghost))
	local near_frame = sync2frame(ns, ghost)
	if not t or not near_frame then return end
	local bi, bv = nil
	for i, f in ipairs(t) do
		local diff = math.abs(f-near_frame)
		if not bv or bv > diff then
			bv = diff
			bi = i
		end
	end
	return frame2sync(t[bi], ghost)
end

function read_pose(info)
	local res = {}
	local big = gd.createFromPng(info[1])
	local i = 0
	local dx, dy = info[3], info[4]
	for line in io.lines(info[2]) do
		local num = tonumber("0x" .. line)
		if num then
			local small = gd.create(2*dx,2*dy)
			gd.copy(small, big, 0, 0, (i % 0x10)*2*dx, math.floor(i/0x10)*2*dy, 2*dx, 2*dy)
			res[num] = small
		end
		i = i+1
	end
	print(string.format("Read %d poses", i))
	return res
end

function get_gfx(ghost)
	return pose_data[ghost.gfx], pose_info[ghost.gfx][3], pose_info[ghost.gfx][4]
end

function draw_own_hitbox()
	local scrollx,scrolly = memory.readword(0x7E0911), memory.readword(0x7E0915)
	local posx, posy = memory.readword(0x7E0AF6), memory.readword(0x7E0AFA)
	local hitx, hity = memory.readword(0x7E0AFE), memory.readword(0x7E0B00)
	local x, y = posx - scrollx, posy - scrolly
	gui.drawbox(x-hitx, y-hity,x+hitx,y+hity,own_color)
end

function draw_ghost_hitbox(ghost,frame)
	local data = ghost.data[frame]
	local scrollx, scrolly =memory.readword(0x7E0911), memory.readword(0x7E0915)
	local x, y = data.posx - scrollx, data.posy - scrolly
	gui.drawbox(x-data.hitx, y-data.hity,x+data.hitx,y+data.hity,ghost.color)
end

function draw_ghost_gfx(ghost,frame)
	if not ghost_gfx then return end
	local data = ghost.data[frame]
	local gfx, dx, dy = get_gfx(ghost)
	local scrollx, scrolly =memory.readword(0x7E0911), memory.readword(0x7E0915)
	local x, y = data.posx - scrollx, data.posy - scrolly
	gui.transparency(2)
	if gfx[data.pose] then gui.gdoverlay(x-dx,y-dy,gfx[data.pose]:gdStr()) end
	gui.transparency(0)
end

function draw_enemy_hitbox()
	local scrollx, scrolly =memory.readword(0x7E0911), memory.readword(0x7E0915)
	for j = 0, 31 do repeat
		local monHP = memory.readword(0x7e0f8c + 0x40*j)
		if monHP == 0 then break end
		local monX, monY = memory.readword(0x7e0f7a + 0x40*j)-scrollx, memory.readword(0x7e0f7e + 0x40*j)-scrolly
		local monRadX, monRadY = memory.readword(0x7e0f82 + 0x40*j), memory.readword(0x7e0f84 + 0x40*j)
		gui.drawbox(monX-monRadX,monY-monRadY,monX+monRadX,monY+monRadY,enemy_color)
	until true end
end

function draw_enemy_hp()
	local scrollx, scrolly =memory.readword(0x7E0911), memory.readword(0x7E0915)
	for j = 0, 31 do repeat
		local monHP = memory.readword(0x7e0f8c + 0x40*j)
		if monHP == 0 then break end
		local monX, monY = memory.readword(0x7e0f7a + 0x40*j)-scrollx, memory.readword(0x7e0f7e + 0x40*j)-scrolly
		local monRadY = memory.readword(0x7e0f84 + 0x40*j)
		gui.text(monX, monY+monRadY, tostring(monHP))
	until true end
end

function draw_shot_hitbox()
	local scrollx, scrolly =memory.readword(0x7E0911), memory.readword(0x7E0915)
	for j = 0,4 do
		local projX, projY = memory.readword(0x7e0b64 + 0x2*j)-scrollx, memory.readword(0x7e0b78 + 0x2*j)-scrolly
		if memory.readword(0x7e0c18 + 0x2*j) >= 0x8000 then
			local w,h = memory.readword(0x7e0bb4 + 0x2*j), memory.readword(0x7e0bc8 + 0x2*j)
			gui.drawbox(projX-w,projY-h,projX+w,projY+h,shot_color)
		end
	end
end

function init_global_map()
	if mini then gui.gdoverlay(map_pos[1],map_pos[2],mini:gdStr())
	else gui.fillbox(map_pos[1], map_pos[2], map_pos[1]+map_pos[3]-1, map_pos[2]+map_pos[4]-1, "black") end
end

function draw_cross(x,y,r,col)
	for i = -r,r do
		gui.drawpixel(x+i,y,col)
		gui.drawpixel(x,y+i,col)
	end
end

function draw_global_map_samus()
	local reg, rx, ry = split_room(get_room())
	local posx, posy = memory.readword(0x7E0AF6), memory.readword(0x7E0AFA)
	local ox, oy = 0, 0
	if region_pos[reg] then
		ox, oy = region_pos[reg].x, region_pos[reg].y
	end
	if reg == 0 and rx >= 49 then rx = rx + 7 end
	local x, y = map_pos[1]+(ox+rx+SHIFT(posx,8))/1, map_pos[2]+(oy+ry+SHIFT(posy,8))/1
	draw_cross(x, y, 1, own_color)
end

function draw_global_map_ghost(ghost, frame)
	local data = ghost.data[frame]
	local reg, rx, ry = split_room(data.room)
	local ox, oy = 0, 0
	if region_pos[reg] then
		ox, oy = region_pos[reg].x, region_pos[reg].y
	end
	if reg == 0 and rx >= 49 then rx = rx + 7 end
	local x, y = map_pos[1]+(ox+rx+SHIFT(data.posx,8))/1, map_pos[2]+(oy+ry+SHIFT(data.posy,8))/1
	draw_cross(x, y, 1, ghost.color)
end

function draw_delay(ghost,delay, index)
	gui.text((index-1)*6*5+1,0,string.format("%d", delay))
	gui.drawline((index-1)*6*5,6,index*6*5-1,6,ghost.color)
	gui.drawline(index*6*5-1,6,index*6*5-1,0,ghost.color)
end

function draw_status()
	local charge = memory.readbyte(0x7e0dc2)
	local spark = memory.readbyte(0x7e0a68)
	local boost = memory.readbyte(0x7e0b3f)
	local speedx = memory.readword(0x7e0b42) + memory.readword(0x7e0b44)/65536
	local speedy = memory.readword(0x7e0b2e) + memory.readword(0x7e0b2c)/65536
	gui.text(0,215,string.format("c:%3d b:%2d s:%3d vx:%7.3f vy:%7.3f", charge, boost, spark, speedx, speedy))
end

-- Not finished. Will not work
function handle_input()
	--local input = input.get()
	--local lmouse = AND(input.mask,0x100) ~= 0
	--if lmouse and not last_lmouse then
	--	-- Switch between ingame and realtime sync
	--	if sync_mode == "ingame" then
	--		sync_mode = "real"
	--	else
	--		sync_mode = "ingame"
	--	end
	--	set_sync(sync_mode)
	--	set_offset(offset_mode)
	--	-- This is slow, and could be cached for about
	--	-- 1 MB memory per ghost.
	--	for i, ghost in pairs(ghosts) do
	--		ghost.syncframes = build_sync(ghost)
	--	end
	--end
	--last_lmouse = lmouse
end

-- End of definitions. Start running.

init()
while true do
	main()
	snes9x.frameadvance()
end

